# coding=utf-8
# Packages automatically included in Anaconda for Python 3.5
# These packages are not added into the Streams application
# bundle when Anaconda is used to build the topology
# https://docs.continuum.io/anaconda/pkg-docs

#
# Packages commented out without any specific details
# are ones that are named in the Anaconda package list
# but do not import as modules using their name

_CONDA_PACKAGES = [
    "alabaster",
    #"anaconda-clean",
    #"anaconda-client",
    #"anaconda-navigator",
    "argcomplete",
    "astropy",
    "babel",
    "bs4",
    "bitarray",
    "blaze",
    "bokeh",
    "boto",
    "bottleneck",
    "cffi",
    "chest",
    "cloudpickle",
    "clyent",
    "colorama",
    "configobj",
    "cryptography",
    "curl",
    "cycler",
    "cython",
    "cytoolz",
    "dask",
    "datashape",
    "decorator",
    "dill",
    "docutils",
    #"dynd-python",
    "et_xmlfile",
    "fastcache",
    "flask",
    #"flask-cors",
    #"fontconfig",
    #"freetype",
    "gevent",
    "greenlet",
    "h5py",
    #"hdf5",
    "heapdict",
    "idna",
    "ipykernel",
    #"ipython",
    "ipython_genutils",
    "ipywidgets",
    "itsdangerous",
    #"jbig",
    "jdcal",
    "jedi",
    "jinja2",
    #"jpeg",
    "jsonschema",
    "jupyter",
    "jupyter_client",
    "jupyter_console",
    "jupyter_core",
    #"libdynd",
    #"libffi",
    #"libgfortran",
    #"libpng",
    #"libsodium",
    #"libtiff",
    #"libxml2",
    #"libxslt",
    "llvmlite",
    "locket",
    "lxml",
    "markupsafe",
    "matplotlib",
    "mistune",
    "mkl",
    "mpmath",
    "multipledispatch",
    "nbconvert",
    "nbformat",
    "networkx",
    "nltk",
    "nose",
    "notebook",
    "numba",
    "numexpr",
    "numpy",
    "odo",
    "openpyxl",
    #"openssl",
    "pandas",
    "partd",
    #"patchelf",
    #"path.py",
    "patsy",
    "pep8",
    "pexpect",
    "pickleshare",
    #"pillow",
    "pip",
    "ply",
    "psutil",
    "ptyprocess",
    "py",
    "pyasn1",
    "pycosat",
    "pycparser",
    #"pycrypto",
    "pycurl",
    "pyflakes",
    "pygments",
    #"pyopenssl",
    "pyparsing",
    #"pyqt",
    #"pytables",
    "pytest",
    #"python",
    #"python-dateutil",
    "pytz",
    #"pyyaml",
    #"pyzmq",
    #"qt",
    # Not in Bluemix "qtawesome",
    "qtconsole",
    "qtpy",
    "readline",
    "redis",
    "requests",
    "rope",
    #"scikit-image",
    #"scikit-learn",
    "scipy",
    "setuptools",
    "simplegeneric",
    "singledispatch",
    "sip",
    "six",
    "snowballstemmer",
    #"sockjs-tornado",
    "sphinx",
    #"spyder",
    "sqlalchemy",
    #"sqlite",
    "statsmodels",
    "sympy",
    "terminado",
    #"tk",
    "toolz",
    "tornado",
    "traitlets",
    "unicodecsv",
    "werkzeug",
    "wheel",
    "xlrd",
    "xlsxwriter",
    "xlwt",
    #"xz",
    "yaml",
    #"zeromq",
    "zlib"
 ]
